"""FeedModifier test cases (with PyTest)."""
from __future__ import annotations

from datetime import datetime, timezone
from math import gcd
from pathlib import Path
from time import sleep

import pytest
import requests
from dateutil import parser
from lxml import etree as ET

from rssreruns.feedmodifier import AtomFeedModifier, FeedModifier, RSSFeedModifier

tmp_dir = Path("test/tmp/")
tmp_dir.mkdir(exist_ok=True)
data_dir = Path("test/feedmodifier/")


def as_RSS(filename: str, **kwargs) -> RSSFeedModifier:
    """Initialize RSSFeedModifier with the given RSS feed."""
    return RSSFeedModifier(data_dir / filename, **kwargs)


def as_Atom(filename: str, **kwargs) -> AtomFeedModifier:
    """Initialize AtomFeedModifier with the given Atom feed."""
    return AtomFeedModifier(data_dir / filename, **kwargs)


@pytest.fixture
def simple_rss_fnames():
    """List of paths for simple RSS feeds."""
    return (
        "no_items.rss",
        "one_item.rss",
        "two_items.rss",
        "four_items.rss",
        "seven_items.rss",
    )


@pytest.fixture
def simple_atom_fnames():
    """List of paths for simple Atom feeds."""
    return (
        "no_items.atom",
        "one_item.atom",
        "two_items.atom",
        "namespaced_two_items.atom",
    )


@pytest.fixture
def simple_fnames(simple_atom_fnames, simple_rss_fnames):
    """List of paths for simple Atom feeds."""
    return (*simple_atom_fnames, *simple_rss_fnames)


@pytest.fixture
def simple_rss_fms(simple_rss_fnames):
    """List of RSSFeedModifiers for simple RSS feeds."""
    return (as_RSS(fname) for fname in simple_rss_fnames)


@pytest.fixture
def simple_atom_fms(simple_atom_fnames):
    """List of AtomFeedModifiers for simple Atom feeds."""
    return (as_Atom(fname) for fname in simple_atom_fnames)


@pytest.fixture
def simple_fms(simple_atom_fms, simple_rss_fms):
    """List of AtomFeedModifiers for simple Atom feeds."""
    return (*simple_atom_fms, *simple_rss_fms)


def test_RSSFeedModifier_init(simple_rss_fms):
    """Test initialization performs correctly for RSSFeedModifiers."""
    for rss_fm in simple_rss_fms:
        assert rss_fm._tree is not None
        assert rss_fm._root is not None
        assert rss_fm._channel is not None


def test_AtomFeedModifier_init(simple_atom_fms):
    """Test initialization performs correctly for AtomFeedModifiers."""
    for atom_fm in simple_atom_fms:
        assert atom_fm._tree is not None
        assert atom_fm._root is not None
        assert atom_fm._channel is not None


def test_from_file(simple_fnames):
    """Test factory method from_file for creating FeedModifiers."""
    for fname in simple_fnames:
        fm = FeedModifier.from_file(data_dir / fname)
        assert fm._tree is not None
        assert fm._root is not None
        assert fm._channel is not None


rss_url = "http://www.rss-specifications.com/blog-feed.xml"
tmp_path = tmp_dir / "my_filename.rss"


# TODO: better way to organize/write this? :(
@pytest.mark.parametrize(
    "args, expected_path",
    [
        # No path specified: expect default
        (
            (rss_url, None),
            "downloads/feed.xml",
        ),
        # Path given as Path object
        ((rss_url, tmp_path), tmp_path),
        # Same path but given as a string ("test/tmp/my-filename.rss")
        ((rss_url, str(tmp_path)), tmp_path),
    ],
)
def test_url_to_file(args, expected_path):
    """Test `url_to_file` with or without a path specified."""
    saved_path = FeedModifier.url_to_file(*args)
    assert saved_path == Path(expected_path)
    assert str(saved_path) == str(expected_path)
    assert saved_path.exists()
    assert saved_path.is_file()

    # Clean up downloaded file and the default /downloads/ directory if it exists
    saved_path.unlink()
    if Path("downloads").exists():
        Path("downloads").rmdir()


# TODO: this is effectively a test for `requests.get`: consider removal?
@pytest.mark.parametrize(
    "bad_url, expected_exception",
    [
        ("just some words", requests.exceptions.MissingSchema),
        ("https://", requests.exceptions.InvalidURL),
    ],
)
def test_url_to_file_invalid_url(bad_url, expected_exception):
    """Test `url_to_file` with an invalid URL argument."""
    with pytest.raises(expected_exception) as exc_info:
        FeedModifier.url_to_file(bad_url)
    assert "Invalid URL" in exc_info.value.args[0]


@pytest.mark.parametrize(
    "url_404",
    [
        "https://example.com/this-will-404-not-a-real-url-34454127.xml",
    ],
)
def test_url_to_file_404(url_404):
    """Test `url_to_file` with a valid URL that will return 404 Not Found."""
    with pytest.raises(ValueError) as exc_info:
        FeedModifier.url_to_file(url_404)
    assert (
        exc_info.value.args[0] == f"Requested url {url_404} returned status code: 404"
    )


@pytest.mark.parametrize(
    "url",
    [rss_url],
)
def test_from_url(url):
    """Test `from_url` with real feed URLs. (Warning: `sleep` used for politeness)."""
    sleep(3)
    fm_via_string = FeedModifier.from_url(url)
    fm_via_file = FeedModifier.from_url(url, path=tmp_dir / "blog-feed.xml")
    assert _same_attributes(fm_via_string, fm_via_file)


def test_from_string(simple_fnames):
    """Test initialization from string rather than url or filepath."""
    for fname in simple_fnames:
        with open(data_dir / fname, "rb") as f:
            contents = f.read()
            # print(type(contents))
            fm = FeedModifier.from_string(contents)

            # Write to file, load from file, and compare
            out_path = tmp_dir / fname
            fm.write(out_path)
            deserialized = FeedModifier.from_file(out_path)
            assert _same_attributes(fm, deserialized)

            # Output as string, load from string, and compare
            out_string = fm.write(out_path)
            deserialized = FeedModifier.from_string(out_string)
            assert _same_attributes(fm, deserialized)


def test_write(simple_fnames, simple_fms):
    """Test writing the modified feed (and associated FeedModifier data) to XML."""
    for fname, fm in zip(simple_fnames, simple_fms):
        out_path = tmp_dir / fname
        fm.write(out_path)
        # TODO: actually write these tests, expected output files, etc.
        deserialized = FeedModifier.from_file(out_path)
        assert _same_attributes(fm, deserialized)
        pass


def _same_attributes(this, other):
    """Pseudo-equality comparison, intended only for testing."""
    # TODO: This is a temporary kludge that only exists for sanity checking. It does
    # not properly show equality, which is why it is not an `__eq__` method.
    # Remove this once better test cases for [de]serialization are in place.
    def _attrs(obj):
        attrs = {
            k: v
            for k, v in vars(obj).items()
            if (k in {"_nsmap"} or not k.startswith("_")) and k not in {"path"}
        }
        from_meta_channel = {
            subelement: obj._meta_channel[subelement].text
            for subelement in (
                "original_title",
                "title_prefix",
                "entry_title_prefix",
                "entry_title_suffix",
                "run_forever",
                "order",
            )
        }
        from_channel = {"title": obj._channel["title"].text}

        # Rewritten without `|` so tox can run tests for Python 3.7
        # return attrs | from_meta_channel | from_channel
        return {**attrs, **from_meta_channel, **from_channel}

    this_attrs = _attrs(this)
    other_attrs = _attrs(other)

    this_nsmap = (
        {(k, v) for k, v in this_attrs["_nsmap"].items()}
        if "_nsmap" in this_attrs
        else set()
    )
    other_nsmap = (
        {(k, v) for k, v in other_attrs["_nsmap"].items()}
        if "_nsmap" in other_attrs
        else set()
    )
    nsmaps_diff = this_nsmap ^ other_nsmap

    this_everything_else = {(k, v) for k, v in this_attrs.items() if k != "_nsmap"}
    other_everything_else = {(k, v) for k, v in other_attrs.items() if k != "_nsmap"}

    diff = this_everything_else ^ other_everything_else
    return len(nsmaps_diff | diff) == 0


def test_same_attributes(simple_fms):
    """Test the `_same_attributes` method (used in testing [de]serialization)."""
    other_fm = as_Atom("two_items.atom")
    other_fm["entry_title_prefix"] = "VERY DIFFERENT PREFIX!"
    for fm in simple_fms:
        assert _same_attributes(fm, fm)
        assert not _same_attributes(fm, other_fm)


def rss_len_examples() -> tuple[RSSFeedModifier, int]:
    """Test cases for RSSFeedModifier's expected number of entries."""
    return [
        (as_RSS("no_items.rss"), 0),
        (as_RSS("one_item.rss"), 1),
        (as_RSS("two_items.rss"), 2),
        (as_RSS("four_items.rss"), 4),
        (as_RSS("seven_items.rss"), 7),
    ]


def atom_len_examples() -> tuple[AtomFeedModifier, int]:
    """Test cases for AtomFeedModifier's expected number of entries."""
    return [
        (as_Atom("no_items.atom"), 0),
        (as_Atom("one_item.atom"), 1),
        (as_Atom("two_items.atom"), 2),
        (as_Atom("namespaced_two_items.atom"), 2),
    ]


@pytest.mark.parametrize(
    "fm, expected_len", (*rss_len_examples(), *atom_len_examples())
)
def test_feed_entries_len(fm, expected_len):
    """Test feed_entries() for FeedModifiers."""
    assert len(fm.feed_entries()) == expected_len


@pytest.mark.parametrize(
    "fm, expected_len", (*rss_len_examples(), *atom_len_examples())
)
def test_feed_meta_entries_len(fm, expected_len):
    """Test _feed_meta_entries() for FeedModifiers."""
    assert len(fm._feed_meta_entries()) == expected_len


@pytest.mark.parametrize(
    "fm, expected_len", (*rss_len_examples(), *atom_len_examples())
)
def test_entries_to_rerun(fm, expected_len):
    """_entries_to_rerun() should initially return all entries."""
    assert len(fm._entries_to_rerun()) == expected_len


@pytest.mark.parametrize(
    "fm, expected_len",
    (
        (fm, length)
        for (fm, length) in (*rss_len_examples(), *atom_len_examples())
        if length > 0
    ),
)
def test_entries_to_rerun_single_rebroadcasts(fm, expected_len):
    """_entries_to_rerun() decreases by one each time rebroadcast(1) is called."""
    assert len(fm._entries_to_rerun()) == expected_len

    # (Go through and rebroadcast all of the entries a few times over.)
    for i in range(7):
        # Rebroadcast entries until there is only 1 that has not been rebroadcasted
        remaining = len(fm._entries_to_rerun())
        while remaining > 1:
            fm.rebroadcast(1)
            assert len(fm._entries_to_rerun()) == remaining - 1
            remaining -= 1

        # Rebroadcast the last remaining entry that has not yet been rebroadcast.
        fm.rebroadcast(1)
        # 0 should not be returned: when rebroadcast(1) would rebroacast the last
        # remaining entry, all entries should be marked as <reran>False</reran>, so
        # the number of remaining entries will again be the total number of entries.
        assert len(fm._entries_to_rerun()) == expected_len


@pytest.mark.parametrize(
    "fm, expected_len, num",
    (
        (fm, length, num)
        for (fm, length) in (*rss_len_examples(), *atom_len_examples())
        for num in range(1, length)
        if length > 0
    ),
)
def test_entries_to_rerun_multiple_rebroadcasts(fm, expected_len, num):
    """Decrease remaining by num (mod len) after calling rebroadcast(num)."""
    assert len(fm._entries_to_rerun()) == expected_len

    # Rewritten with `gcd` because Python 3.7 lacks `math.lcm` (but has `gcd`)
    #   lcm = math.lcm(expected_len, num)
    lcm = abs(expected_len * num) // gcd(expected_len, num)

    calls_to_reach_exactly_zero = lcm // num
    # print(f"Expected_len: {expected_len}, num: {num}")
    # print(f"LCM: {lcm}")
    # print(f"(lcm // num): {(lcm // num)}")
    # print(f"(lcm // num) - 1: {(lcm // num) - 1}")

    # (Go through and rebroadcast all of the entries a few times over.)
    for i in range(7):
        remaining = expected_len
        # print(f"{remaining} (mod {expected_len})")
        for calls_to_rebroadcast in range(calls_to_reach_exactly_zero - 1):
            fm.rebroadcast(num)
            now_remaining = (remaining - num) % expected_len
            assert len(fm._entries_to_rerun()) % expected_len == now_remaining
            remaining = now_remaining
            # print(f"{remaining} (mod {expected_len})")

        # After (lcm // num) - 1 rebroadcasts of num entries, there should be exactly
        # `num` entries left to rebroadcast
        assert remaining == num

        # Rebroadcast the last `num` remaining entries
        fm.rebroadcast(num)
        # 0 should not be returned: when rebroadcast(num) would rebroacast the last
        # remaining entries, all entries should be marked as <reran>False</reran>, so
        # the number of remaining entries will again be the total number of entries.
        assert len(fm._entries_to_rerun()) == expected_len


def test_update_entry_pubdate(simple_fms):
    """Test `update_entry_pubdate` for FeedModifiers."""
    for fm in simple_fms:
        num_entries = len(fm.feed_entries())
        elements_dates = []
        for entry in fm.feed_entries():
            dt_now = datetime.now(timezone.utc)
            updated = fm.update_entry_pubdate(entry, dt_now)
            elements_dates.append((updated, dt_now))

        # The number of entries should remain the same
        assert len(fm.feed_entries()) == num_entries

        for elements, dt in elements_dates:
            # Check that the correct elements were updated (different depending on
            # type of feed)
            if isinstance(fm, RSSFeedModifier):
                # An RSSFeedModifier only updates `pubDate`
                assert len(elements) == 1
                assert elements[0].tag.rpartition("}")[2] == "pubDate"
            else:
                # An AtomFeedModifier should update both `published` and `updated`
                assert len(elements) == 2
                assert {el.tag.rpartition("}")[2] for el in elements} == {
                    "published",
                    "updated",
                }

            # Check that the datetimes were set as expected
            for el in elements:
                assert el.text == fm.format_datetime(dt)


def test_RSSFeedModifier_format_datetime_simple():
    """Test formatting of datetimes to strings for RSSFeedModifier."""
    example_datetimes = [
        "Mon, 15 Mar 2021 14:32:20 -0400",
        "Sun, 19 May 2002 15:21:36 +0000",
    ]

    for dt in example_datetimes:
        dt_parsed = parser.parse(dt)
        dt_formatted = RSSFeedModifier.format_datetime(dt_parsed)
        assert dt == dt_formatted


def test_AtomFeedModifier_format_datetime_simple():
    """Test formatting of datetimes to strings for AtomFeedModifiers.

    Adapted from https://datatracker.ietf.org/doc/html/rfc4287#section-3.3
    """
    example_datetimes = [
        "2003-12-13T18:30:02Z",
        "2003-12-13T18:30:02.700000Z",
        "2003-12-13T18:30:02+01:00",
        "2003-12-13T18:30:02.700000+01:00",
    ]

    for dt in example_datetimes:
        dt_parsed = parser.parse(dt)
        dt_formatted = AtomFeedModifier.format_datetime(dt_parsed)
        assert dt == dt_formatted


def test_source_url_simple(simple_rss_fms, simple_atom_fms):
    """Test source_url() returns the correct source url.

    For RSS feeds, the feed channel has a single link element.

    For Atom feeds, the feed may have multiple link elements, but should have one
    with rel="self" (which will be returned if present.)
    """
    for fm in simple_rss_fms:
        assert fm.source_url() == "http://example.com/notarealblogurl.htm"

    for fm in simple_atom_fms:
        assert fm.source_url() == "https://example.org/feed.atom"


def test_source_url_xml_base():
    """Test source_url() resolves URIs relative to xml:base correctly when needed.

    See:
        https://www.w3.org/TR/xmlbase/
        https://feedparser.readthedocs.io/en/latest/resolving-relative-links.html
    """
    # Base URI https://www.example.org/ declared in the root
    fm = as_Atom("namespaced_xml_base_in_root.atom")
    assert fm.source_url() == "https://www.example.org/blog/news/feed.atom"

    # Same base URI declared in the root, and "blog/news/" declared in the <link/>
    # itself
    fm = as_Atom("namespaced_xml_base_in_root_and_link.atom")
    assert fm.source_url() == "https://www.example.org/blog/news/feed.atom"


def test_feed_type(simple_rss_fms, simple_atom_fms):
    """Test feed_type() returns "RSS" or "Atom" as expected."""
    # Base URI https://www.example.org/ declared in the root
    for fm in simple_rss_fms:
        assert fm.feed_type() == "RSS"

    for fm in simple_atom_fms:
        assert fm.feed_type() == "Atom"


def w3c_feed_validator(path: Path | str) -> bool:
    """Validate an Atom/RSS feed through the W3C's public validation service."""
    base_uri = "http://validator.w3.org/feed/check.cgi"
    # At LEAST 1 second delay between requests is required
    sleep(2)

    with open(path, "r") as f:
        feed = f.read()

    response = requests.post(
        base_uri,
        params={"rawdata": feed, "output": "soap12"},
        headers={"Content-type": "application/xml"},
    )
    print(response)
    print(response.content)
    print("============================")
    response_file = "w3c_out.xml"
    with open(response_file, "wb") as f:
        f.write(response.content)
    tree = ET.parse(response_file)
    tree.write(response_file, pretty_print=True, xml_declaration=True, encoding="utf-8")
    return True


if False:
    w3c_feed_validator(data_dir / "namespaced_two_items.atom")



"""Modify a given XML-based feed (RSS or Atom)."""
from __future__ import annotations

import copy
import email.utils
import random
import uuid
from abc import ABC, abstractmethod
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Optional

import requests
from dateutil import parser
from lxml import etree as ET

from .elementwrapper import ElementWrapper, ElementWrapperFactory

# Underscores present because those are the class names in lxml.etree
Element = ET._Element
ElementTree = ET._ElementTree
QName = ET.QName


class FeedModifier(ABC):
    """Modify a given XML-based feed.

    Concrete subclasses correspond to the specific kind of feed, i.e. RSS or Atom.
    """

    _bool_attributes = {"run_forever", "overwrite_entry_ids"}

    def __init__(
        self,
        path: str | Path | None,
        contents: str | bytes | None = None,
        schedule_kwargs: Optional[Any] = None,
        run_forever: Optional[bool] = None,
        initialize_all_pubdates: bool = False,
        overwrite_self_link: str | None = None,
        overwrite_entry_ids: bool | None = None,
        title_kwargs: dict[str, Any] = {},
        entry_title_kwargs: dict[str, str] = {},
    ) -> None:
        """Initialization."""
        if path is None and contents is None:
            raise ValueError("Arguments path and contents cannot both be None.")

        self.path = Path(path) if path is not None else None

        self._tree: ElementTree = self._parse_file_or_string(
            path=path, contents=contents
        )

        # Prefix and URI for the `reruns` XML namespace
        self._ns_prefix = "reruns"
        self._ns_uri = "https://github.com/hannahlog/rss-reruns"

        # Local names of elements containing feed and entry data
        self._meta_channel_tag = f"{self._ns_prefix}:channel_data"
        self._meta_entry_tag = f"{self._ns_prefix}:entry_data"

        # Declaration added to the root element as
        #   `xmlns:reruns="https://github.com/hannahlog/rss-reruns"`
        self.add_namespace(prefix=self._ns_prefix, uri=self._ns_uri)

        self._default_EWF = ElementWrapperFactory("" if "" in self._nsmap else None)
        self._root: ElementWrapper = self._default_EWF(self._tree.getroot())

        # Element containing metadata and entry/item elements:
        # `feed` for Atom (which is also the root), `channel` for RSS (not the root)
        self._channel: ElementWrapper = self.feed_channel()

        # Subelement of the channel containing data and settings for the FeedModifier
        self._meta_channel: ElementWrapper = self._channel[self._meta_channel_tag]

        # Default meta channel values:
        meta_channel_defaults = {
            "order": "chronological",
            "rate": "1",
            "run_forever": "True",
            "original_title": self._channel["title"].text or "",
            "original_self_link": self.source_url(),
            "overwrite_entry_ids": "False",
            "title_prefix": None,
            "title_suffix": None,
            "entry_title_prefix": None,
            "entry_title_suffix": None,
        }
        self._create_defaults_if_missing(self._meta_channel, meta_channel_defaults)

        # Default meta entry values:
        for entry in self.feed_entries():
            entry_meta = entry[self._meta_entry_tag]
            meta_entry_defaults = {
                "original_pubdate": self.get_entry_pubdate(entry),
                "original_title": entry["title"].text or "",
                "reran": "False",
            }
            self._create_defaults_if_missing(entry_meta, meta_entry_defaults)
            if initialize_all_pubdates:
                self.update_entry_pubdate(entry, datetime.now(timezone.utc))

        if run_forever is not None:
            self["run_forever"] = run_forever

        if overwrite_entry_ids is not None:
            self["overwrite_entry_ids"] = overwrite_entry_ids

        if overwrite_self_link:
            self._overwrite_self_link(new_self_link=overwrite_self_link)

        self.set_feed_title(**title_kwargs)
        self.set_entry_titles(**entry_title_kwargs)

    def _clean_nsmap(self, nsmap: dict[Optional[str], str]) -> dict[str, str]:
        """Replace `None` keys with empty strings."""
        return {(k or ""): v for k, v in nsmap.items()}

    def add_namespace(self, prefix: str, uri: str) -> None:
        """Add the given namespace to the root element (if not already present)."""
        nsmap = {k: v for k, v in self._tree.getroot().nsmap.items() if k}
        nsmap[prefix] = uri

        # All pre-existing namespaces are kept.
        # If a default namespace was present, it will be kept, even though it
        # is not included here in nsmap or all_prefixes.
        all_prefixes = list(nsmap)
        ET.cleanup_namespaces(
            self._tree, top_nsmap=nsmap, keep_ns_prefixes=all_prefixes
        )
        self._nsmap = self._clean_nsmap(self._tree.getroot().nsmap)
        pass

    def _create_defaults_if_missing(
        self, parent: ElementWrapper, defaults: dict[str, Optional[str]]
    ) -> None:
        """Create subelements with given text only if the subelement does not exist."""
        for tag, text in defaults.items():
            # Defaults do not override already-existing elements or their texts.
            # Specifically, if the subelement already exists with no text (<tag/>),
            # then this will preserve the subelement having no text.
            # E.g. a modified feed that previously had entry_title_prefix set
            # to None will remain that way:
            #   `<reruns:entry_title_prefix/>`
            # Preseving such None-text is why it is written this way, and not
            #   `parent[tag].text = parent[tag].text or text`
            # which would overwrite "" as well as None.
            if tag not in parent:
                parent[tag].text = text

    @classmethod
    def _parse_file_or_string(cls, path=None, contents=None) -> ElementTree:
        """Parse from filepath or string."""
        # If blank text is preserved, then `ET.write` will not properly indent
        # any newly added elements which contain text, even with pretty printing.
        # However, if the parser removes blank text, then `write()` will add new
        # indentation to the entire document, so all elements wil be properly indented
        # (if pretty printing is enabled.)
        #
        # See: "Why doesn't the pretty_print option reformat my XML output?"
        # lxml.de/FAQ.html#why-doesn-t-the-pretty-print-option-reformat-my-xml-output
        parser = ET.XMLParser(remove_blank_text=True, strip_cdata=False)
        if path is not None:
            # base_url="" is needed to prevent lxml from using the document's filepath
            # as an implicit base URI if no `xml:base` is declared
            return ET.parse(path, parser=parser, base_url="")
        elif contents is not None:
            return ET.fromstring(contents, parser=parser).getroottree()
        else:
            raise ValueError("Either path or contents must not be None.")

    @classmethod
    def from_url(cls, url, *, path=None, feed_format=None, **kwargs) -> "FeedModifier":
        """Create a FeedModifier from a given source feed's url."""
        if path is not None:
            saved_path = cls.url_to_file(url, path)
            kwargs.pop("url", None)
            return cls.from_file(saved_path, **kwargs)
        else:
            contents = cls.url_to_contents(url)
            kwargs.pop("url", None)
            return cls.from_string(contents, **kwargs)

    @classmethod
    def from_string(
        cls, contents: str | bytes, *, feed_format=None, **kwargs
    ) -> "FeedModifier":
        """Create a FeedModifier from a given source feed's string."""
        if feed_format is not None:
            concrete_subclass = (
                RSSFeedModifier if "rss" in feed_format.lower() else AtomFeedModifier
            )
        else:
            concrete_subclass = cls._infer_format_from_contents(
                path=None, contents=contents
            )

        kwargs.pop("feed_format", None)
        kwargs.pop("path", None)
        kwargs.pop("contents", None)
        return concrete_subclass(path=None, contents=contents, **kwargs)

    @classmethod
    def from_file(cls, path, *, feed_format=None, **kwargs) -> "FeedModifier":
        """Create a FeedModifier from a given source feed's path."""
        if feed_format is not None:
            concrete_subclass = (
                RSSFeedModifier if "rss" in feed_format.lower() else AtomFeedModifier
            )
        else:
            concrete_subclass = cls._infer_format(path)

        kwargs.pop("feed_format", None)
        return concrete_subclass(path, **kwargs)

    @classmethod
    def _infer_format(cls, path: str | Path) -> type["FeedModifier"]:
        """Guess the format, RSS or Atom, of a given feed file."""
        path = Path(path)
        if not (path.exists() and path.is_file()):
            raise ValueError(f"Given path does not refer to a feed file: {path}")

        # Trust the file extension if it specifies .rss or .atom
        if ".rss" in [suffix.lower() for suffix in path.suffixes]:
            return RSSFeedModifier
        elif ".atom" in [suffix.lower() for suffix in path.suffixes]:
            return AtomFeedModifier

        # Otherwise, inspect the actual contents
        try:
            return cls._infer_format_from_contents(path=path)
        except ValueError as e:
            raise ValueError(f"Format of file {path} could not be determined.") from e

    @classmethod
    def _infer_format_from_contents(
        cls, path=None, contents=None
    ) -> type["FeedModifier"]:
        """Guess the format, RSS or Atom, of a given feed from its string contents."""
        # TODO: Wasteful to parse the entire file just to infer the feed's
        # format -- find less wasteful solution?
        # (To just check the root element?)
        root: Element = cls._parse_file_or_string(
            path=path, contents=contents
        ).getroot()

        if "rss" in root.tag.lower():
            return RSSFeedModifier
        elif "feed" in root.tag.lower():
            return AtomFeedModifier
        else:
            raise ValueError(
                f"Format of feed could not be determined. " f"Root element: {root.tag}"
            )

    @staticmethod
    def url_to_file(url: str, path: Optional[str | Path] = None) -> Path:
        """Download and save XML from given URL."""
        path = Path(path or "downloads/feed.xml")
        response = requests.get(url)
        if not response.ok:
            raise ValueError(
                f"Requested url {url} returned status code: {response.status_code}"
            )

        path.parents[0].mkdir(exist_ok=True, parents=True)
        with open(path, "wb") as f:
            f.write(response.content)
        return path

    @staticmethod
    def url_to_contents(url: str) -> bytes:
        """Download XML from given URL."""
        response = requests.get(url)
        if not response.ok:
            raise ValueError(
                f"Requested url {url} returned status code: {response.status_code}"
            )
        return response.content

    def __getitem__(self, name) -> str:
        """Access meta channel subelements as if they're items of the FeedModifier."""
        if name in self._meta_channel:
            text = self._meta_channel[name].text
        elif name in self._channel:
            text = self._channel[name].text
        else:
            raise KeyError(name)
        return text if name not in self._bool_attributes else bool(text)

    def __setitem__(self, name, value):
        """Access meta channel subelements as if they're items of the FeedModifier."""
        if name in self._bool_attributes and str(value).capitalize() not in {"True", "False"}:
            raise ValueError(
                f"Invalid value {value} for attribute {name}: expected True or False."
            )
        self._meta_channel[name].text = (
            value if name not in self._bool_attributes else str(value).capitalize()
        )
        pass

    def __delitem__(self, name):
        """Access meta channel subelements as if they're items of the FeedModifier."""
        del self._meta_channel[name]
        pass

    def set_feed_title(
        self,
        *,
        prefix: str | None = None,
        suffix: str | None = None,
    ) -> str:
        """Specify the new feed's title by a prefix and suffix string.

        TODO: Functionality overhauled: rewrite docstring.

        The new title may be specified through exactly one of the keyword arguments:
        `title` to give an exact string, `prefix` to prepend a string to the original
        title, or `func` to provide a function that, given the old title as a string,
        creates a new title string.

        Args:
            prefix (str | None):
                String to prepend to the original title.
            suffix (str | None):
                Suffix to append to the original title.

        Returns:
            str: the title of the republished feed, as it will be written to file.
        """
        # Set the new prefix and suffix strings if given
        if prefix:
            self["title_prefix"] = prefix
        if suffix:
            self["title_suffix"] = suffix

        new_title_list: list[str] = [
            self._meta_channel[tag].text
            for tag in (
                "title_prefix",
                "original_title",
                "title_suffix",
            )
            if tag in self._meta_channel and self._meta_channel[tag].text is not None
        ]
        self._channel["title"].text = " ".join(new_title_list)
        return self._channel["title"]

    def set_entry_titles(
        self, prefix: Optional[str] = None, suffix: Optional[str] = None
    ):
        """Set the entry titles."""
        # Set the new prefix and suffix strings if given
        if prefix:
            self["entry_title_prefix"] = prefix
        if suffix:
            self["entry_title_suffix"] = suffix
        for entry in self.feed_entries():
            # TODO: This is unacceptably sloppy. Organize and make this readable.
            # Consider refactoring somehow.
            meta_entry = entry[self._meta_entry_tag]

            # Initialize dict of the title parts that exist
            tags = ("entry_title_prefix", "original_title", "entry_title_suffix")
            parents = (self._meta_channel, meta_entry, self._meta_channel)
            title_parts: dict[str, str] = {
                tag: parent[tag].text
                for tag, parent in zip(tags, parents)
                if tag in parent and parent[tag].text
            }

            # Apply the original date to the prefix and/or suffix if the original date
            # is available
            original_date = meta_entry["original_pubdate"].text
            if original_date is not None:
                dt = parser.parse(original_date)
                affixes = {"entry_title_prefix", "entry_title_suffix"}.intersection(
                    title_parts
                )
                for affix in affixes:
                    title_parts[affix] = dt.strftime(title_parts[affix])

            title_list: list[str] = list(title_parts.values())
            entry["title"].text = " ".join(title_list)
        pass

    def num_remaining(self) -> int:
        """Number of entries that have not yet been rebroadcast."""
        return len(self._entries_to_rerun())

    def _entries_to_rerun(self) -> list[ElementWrapper]:
        """Entries that have not yet been rebroadcast."""
        not_reran = [
            self._default_EWF(meta_entry.getparent())
            for meta_entry in self._feed_meta_entries()
            if meta_entry["reran"].text.lower() == "false"
        ]

        if len(not_reran) == 0 and self["run_forever"]:
            self._mark_all_not_reran()
            return self.feed_entries()
        else:
            return not_reran

    def _feed_meta_entries(self) -> list[ElementWrapper]:
        """Returns iterator over the meta subelements of the feed's entries."""
        return [entry[self._meta_entry_tag] for entry in self.feed_entries()]

    def _mark_all_not_reran(self) -> None:
        """Mark all entries as not having been rebroadcast yet."""
        for meta_entry in self._feed_meta_entries():
            meta_entry["reran"].text = "False"
        pass

    def rebroadcast(
        self, num: int = 1, use_datetime: Optional[datetime | str] = None
    ) -> list[ElementWrapper]:
        """Update the publication date for the given number of entries."""
        if num < 0:
            raise ValueError(f"Cannot select negative number of entries: {num}")

        reran = []
        remaining = num
        while remaining > 0:
            entries = self._entries_to_rerun()
            if remaining >= len(entries):
                for entry in entries:
                    self._rebroadcast_entry(entry, use_datetime)
                reran += entries
                remaining -= len(entries)
            else:
                if self._meta_channel["order"].text.lower() == "chronological":
                    entries.sort(key=self.get_entry_original_pubdate)
                else:
                    random.shuffle(entries)
                for i in range(remaining):
                    self._rebroadcast_entry(entries[i], use_datetime)
                reran += entries[0:remaining]
                remaining = 0
        return reran

    def _rebroadcast_entry(
        self, entry: ElementWrapper, use_datetime: Optional[datetime | str] = None
    ) -> None:
        """AAAAAAAAAAAA."""
        if use_datetime is None:
            dt = datetime.now(timezone.utc)
        else:
            dt = (
                use_datetime
                if isinstance(use_datetime, datetime)
                else parser.parse(use_datetime)
            )
        self.update_entry_pubdate(entry, dt)
        self._update_entry_uuid(entry)
        entry[self._meta_entry_tag]["reran"].text = "True"

    def write(
        self,
        path: str | Path | None,
        with_reruns_data: bool = True,
        pretty_print: bool = True,
        use_datetime: datetime | str | None = None,
    ) -> str:
        """Write modified feed (RSS or Atom) to XML file."""
        # Update when the feed itself was last built before writing out
        if use_datetime is None:
            dt = datetime.now(timezone.utc)
        else:
            dt = (
                use_datetime
                if isinstance(use_datetime, datetime)
                else parser.parse(use_datetime)
            )
        self.update_feed_builddate(dt)

        # If reruns metadata is included, write out the original tree
        tree_to_write = self._tree
        if not with_reruns_data:
            # Otherwise, make a stripped copy of the tree
            stripped_tree = copy.deepcopy(self._tree)
            stripped_root = stripped_tree.getroot()

            # Remove `reruns` elements from the tree's copy
            ET.strip_elements(
                stripped_tree,
                self._meta_entry_tag.split(":")[1],
                self._meta_channel_tag.split(":")[1],
            )

            # Remove `reruns` namespace declaration
            nsmap: dict[str, str] = {
                k: v
                for k, v in stripped_root.nsmap.items()
                if k is not None and k != self._ns_prefix
            }
            ET.cleanup_namespaces(
                stripped_tree, top_nsmap=nsmap, keep_ns_prefixes=list(nsmap)
            )
            tree_to_write = stripped_tree

        if path is not None:
            tree_to_write.write(
                path,
                pretty_print=pretty_print,
                xml_declaration=True,
                encoding="utf-8",
            )
        return ET.tounicode(
            tree_to_write,
            pretty_print=pretty_print,  # xml_declaration=True, encoding="utf-8"
        )

    def to_string(self, **kwargs) -> str:
        """Output the tree as a string. See `write` for optional kwargs."""
        return self.write(path=None, **kwargs)

    def get_entry_original_pubdate(self, entry: ElementWrapper) -> datetime:
        """Get a given entry/item's original date of publication."""
        original_date = entry[self._meta_entry_tag]["original_pubdate"].text
        if original_date is None:
            raise ValueError("Entry missing original_pubdate")
        return parser.parse(original_date)

    def feed_type(self) -> str:
        """Returns the feed format, "RSS" or "Atom".

        NOTE: Relies on the class name of the concrete instance. If more subclasses
        are made without 'RSS' or 'Atom' in the class name, for some other XML-based
        syndication format (lol), that subclass should override this method.
        """
        if "rss" in self.__class__.__name__.lower():
            return "RSS"
        elif "atom" in self.__class__.__name__.lower():
            return "Atom"
        else:
            raise RuntimeError(
                f"Feed format of {self} instance could not be determined."
            )

    def _absolute_url(self, entry: ElementWrapper) -> str:
        """Resolve a relative url and its base url into an absolute url."""
        return entry.href if not entry.base else "".join([entry.base, entry.href])

    def _generate_entry_uuid(self, entry: ElementWrapper) -> str:
        """Generate a UUID for a given entry.

        The generated UUID will not be the same as the atom:id or rss:guid for the entry
        (if one is already present.)

        It may not technically be "correct" to replace an item's existing uuid if one
        considers the republished entry to be "the same" as the original. Unfortunately,
        some feed readers cache the entry's information (including publication date)
        with the uuid as the key -- as a result, the reader won't recognize that an
        entry has had its publication date changed if the entry's uuid remains the same.
        """
        return uuid.uuid4().urn

    @abstractmethod
    def _update_entry_uuid(self, entry: ElementWrapper) -> ElementWrapper:
        """Update the entry's uuid (the guid or id for RSS or Atom respectively).

        Returns the element (<guid> or <id>) containing the new uuid.
        """
        pass

    @abstractmethod
    def feed_channel(self) -> ElementWrapper:
        """Returns the `feed` (Atom) or `channel` (RSS) element of the tree."""
        pass

    @abstractmethod
    def feed_entries(self) -> list[ElementWrapper]:
        """Returns iterator over the feed's entry/item elements."""
        pass

    @abstractmethod
    def get_entry_pubdate(self, entry: ElementWrapper) -> Optional[str]:
        """Get a given entry/item's date of publication."""
        pass

    @abstractmethod
    def update_entry_pubdate(
        self, entry: ElementWrapper, date: datetime
    ) -> list[ElementWrapper]:
        """Update a given entry/item's date of publication."""
        pass

    @abstractmethod
    def update_feed_builddate(self, date: datetime) -> list[ElementWrapper]:
        """Update the feed's datetime of last publication."""
        pass

    @abstractmethod
    def source_url(self) -> str:
        """Return the original feed's url."""
        pass

    @abstractmethod
    def _overwrite_self_link(self, new_self_link: str) -> None:
        """Create or write over the url of the `<atom:link rel='self'/>` element."""
        pass

    @staticmethod
    @abstractmethod
    def format_datetime(date: datetime) -> str:
        """Format a datetime as a string, in the format to be written to file."""
        pass


class RSSFeedModifier(FeedModifier):
    """Modify a given RSS feed."""

    def feed_channel(self) -> ElementWrapper:
        """Returns the `feed` (Atom) or `channel` (RSS) element of the tree."""
        # For RSS, the `channel` element is a child of the root `rss` element
        channel = self._default_EWF(self._root)["channel"]
        if channel is None:
            raise ValueError("RSS feed must contain `channel` element (not found).")
        else:
            return channel

    def feed_entries(self) -> list[ElementWrapper]:
        """Returns iterator over the feed's item elements."""
        return self._channel.iterfind("item")

    def get_entry_pubdate(self, entry: ElementWrapper) -> Optional[str]:
        """Get a given entry/item's date of publication."""
        pubdate = entry["pubDate"]
        if pubdate is not None:
            return pubdate.text
        raise ValueError(f"RSS entry has no 'pubdate': {entry._element}")

    def update_entry_pubdate(
        self, entry: ElementWrapper, date: datetime
    ) -> list[ElementWrapper]:
        """Update a given entry/item's date of publication."""
        formatted_date: str = self.format_datetime(date)
        entry["pubDate"].text = formatted_date
        return [entry["pubDate"]]

    def update_feed_builddate(self, date: datetime) -> list[ElementWrapper]:
        """Update the feed's datetime of last publication."""
        # For RSS, this updates the channel's `lastBuildDate` and `pubDate`
        formatted_date: str = self.format_datetime(date)
        self._channel["pubDate"].text = formatted_date
        self._channel["lastBuildDate"].text = formatted_date
        return [self._channel["pubDate"], self._channel["lastBuildDate"]]

    def source_url(self) -> str:
        """Return the original feed's url."""
        return self._channel["link"].text

    def _overwrite_self_link(self, new_self_link: str) -> None:
        """Create or write over the url of the `<atom:link rel='self'/>` element."""
        if "atom" not in self._nsmap:
            self.add_namespace(prefix="atom", uri="http://www.w3.org/2005/Atom")

        # Per w3c recommendations for RSS:
        # https://validator.w3.org/feed/docs/warning/MissingAtomSelfLink.html
        atom_link = self._channel["atom", "link"]
        self._channel["link"].addprevious(atom_link._element)
        atom_link.clear_xml_base()
        atom_link.href = new_self_link
        atom_link.rel = "self"
        atom_link.type = "application/rss+xml"

    def _update_entry_uuid(self, entry: ElementWrapper) -> ElementWrapper:
        """Update the entry's uuid in the <guid> subelement.

        Returns the subelement containing the new uuid.
        """
        entry["guid"].text = self._generate_entry_uuid(entry)
        entry["guid"].isPermaLink = "false"
        return entry["guid"]

    @staticmethod
    def format_datetime(date: datetime) -> str:
        """Format a datetime as a string, in the format required for RSS.

        The RSS 2.0 specification stipulates:

        "All date-times in RSS conform to the Date and Time Specification of RFC 822,
        with the exception that the year may be expressed with two characters or four
        characters (four preferred)."
        (https://www.rssboard.org/rss-specification)

        The functions `formatdate` and `format_datetime` in `emails.util` conform to
        RFC 2822, which means their datetimes conform to RFC 822.
        (https://docs.python.org/3/library/email.utils.html#email.utils.format_datetime)

        `format_datetime`is used below for our purposes.

        Args:
            date (datetime):
                Date to be formatted.

        Returns:
            str: correctly-formatted string representing the datetime.
        """
        return email.utils.format_datetime(date)


class AtomFeedModifier(FeedModifier):
    """Modify a given Atom feed."""

    def feed_channel(self) -> ElementWrapper:
        """Returns the `feed` (Atom) or `channel` (RSS) element of the tree."""
        # For Atom, the `feed` element is the root itself
        return self._default_EWF(self._root)

    def feed_entries(self) -> list[ElementWrapper]:
        """Returns iterator over the feed's entry elements."""
        return self._default_EWF(self._root).iterfind("entry")

    def get_entry_pubdate(self, entry: ElementWrapper) -> Optional[str]:
        """Get a given entry/item's date of publication."""
        if "updated" in entry:
            return entry["updated"].text
        elif "published" in entry:
            return entry["published"].text
        raise ValueError(
            f"Atom entry has no 'updated' nor 'published': {entry._element}"
        )

    def update_entry_pubdate(
        self, entry: ElementWrapper, date: datetime
    ) -> list[ElementWrapper]:
        """Update a given entry/item's date of publication."""
        formatted_date: str = self.format_datetime(date)
        entry["published"].text = formatted_date
        entry["updated"].text = formatted_date
        return [entry["published"], entry["updated"]]

    def update_feed_builddate(self, date: datetime) -> list[ElementWrapper]:
        """Update the feed's datetime of last publication."""
        # For Atom, this updates the channel's `updated` element
        formatted_date: str = self.format_datetime(date)
        self._channel["updated"] = formatted_date
        return [self._channel["updated"]]

    def source_url(self) -> str:
        """Return the original feed's url."""
        if "original_self_link" in self._meta_channel:
            return self._meta_channel["original_self_link"].text
        links = self._channel.iterfind("link")
        # An Atom feed may have multiple link elements, and "SHOULD" have one
        # link element with the attribute `rel="self"`.
        # (See RFC 4287:
        # https://datatracker.ietf.org/doc/html/rfc4287#section-4.1.1)
        self_links = [link for link in links if link.rel == "self"]
        if links:
            link = (self_links or links)[0]
        else:
            # TODO: not actually a ValueError
            raise ValueError("Atom feed contains no feed link element.")

        # The only reason to not just immediately return
        #
        #    (self_links or links)[0].href
        #
        # in the above is edge cases resolving uris relative to xml:base
        return self._absolute_url(link)

    def _overwrite_self_link(self, new_self_link: str) -> None:
        """Create or write over the url of the `<atom:link rel='self'/>` element."""
        links = self._channel.iterfind("link")
        self_links = [link for link in links if link.rel == "self"]
        self_link = (
            self_links[0] if self_links else self._channel.create_subelement("link")
        )
        if not self_links:
            links[0].addprevious(self_link._element)

        self_link.clear_xml_base()
        self_link.href = new_self_link
        self_link.rel = "self"
        self_link.type = "application/atom+xml"

    def _update_entry_uuid(self, entry: ElementWrapper) -> ElementWrapper:
        """Update the entry's uuid in the <id> subelement.

        Returns the subelement containing the new uuid.
        """
        entry["id"].text = self._generate_entry_uuid(entry)
        entry["id"].isPermaLink = "false"
        return entry["id"]

    @staticmethod
    def format_datetime(date: datetime) -> str:
        """Format a datetime as a string, in the format required for Atom.

        The Atom specification stipulates:

        "A Date construct is an element whose content MUST conform to the "date-time"
        production in [RFC3339].  In addition, an uppercase "T" character MUST be used
        to separate date and time, and an uppercase "Z" character MUST be present in the
        absence of a numeric time zone offset. [...]

        Such date values happen to be compatible with the following specifications:
        [ISO.8601.1988], [W3C.NOTE-datetime-19980827], and
        [W3C.REC-xmlschema-2-20041028]."

        (https://datatracker.ietf.org/doc/html/rfc4287#section-3.3)

        Args:
            date (datetime):
                Date to be formatted.

        Returns:
            str: correctly-formatted string representing the datetime.
        """
        return date.isoformat("T").replace("+00:00", "Z")


if __name__ == "__main__":
    pass



"""ElementWrapper test cases (with PyTest)."""
from __future__ import annotations

from pathlib import Path

from lxml import etree as ET

from rssreruns.elementwrapper import ElementWrapper, ElementWrapperFactory

tmp_dir = Path("test/tmp/")
tmp_dir.mkdir(exist_ok=True)
data_dir = Path("test/elementwrapper/")

parser = ET.XMLParser(remove_blank_text=True)


def test_elementwrapper_getitem_existing_subelements():
    """Basic functionality of __getitem__ to access existing subelements."""
    html = "<html><head/><body><h1>Heading</h1><p>Hello!</p><hr/></body></html>"
    html_root = ET.XML(html)
    html_tree = html_root.getroottree()

    wrapped_root = ElementWrapper(html_root, None)
    assert wrapped_root._element.tag == "html"
    assert wrapped_root._element.text is None

    wrapped_body = wrapped_root["body"]
    assert wrapped_body._element.tag == "body"
    assert wrapped_body._element.text is None

    wrapped_p = wrapped_root["body"]["p"]
    also_wrapped_p = wrapped_body["p"]
    assert wrapped_p._element.tag == "p"
    assert also_wrapped_p._element.tag == "p"
    assert wrapped_p._element.text == "Hello!"

    # Sanity check that __getitem__ has not *modified* the tree.
    html_after = ET.tounicode(html_tree)
    assert html == html_after


def test_elementwrapper_getitem_default_namespace():
    """Functionality of __getitem__ to access default-namespaced subelements."""
    tree = ET.parse(data_dir / "default_namespace.xml", parser=parser)
    root = tree.getroot()
    xml_before = ET.tounicode(tree)

    # From <book xmlns='urn:loc.gov:books' xmlns:isbn='urn:ISBN:0-395-36341-6'>:
    # Default namespace
    default_prefix = ""
    default_uri = "urn:loc.gov:books"
    # Other namespace
    isbn_prefix = "isbn"
    isbn_uri = "urn:ISBN:0-395-36341-6"

    # Factory for ElementWrappers in the default namespace
    default_EWF = ElementWrapperFactory(default_prefix)

    # Subelement access will use the default namespace, unless another namespace prefix
    # or URI is given explicitly
    wrapped_root_default = default_EWF(root)
    assert wrapped_root_default._with_prefix == ""

    # Subelements in the default namespace should be accessible with or without the
    # prefix (or URI) given explicitly
    for arg in [
        "title",
        ("", "title"),
        f"{{{default_uri}}}title",
        (default_uri, "title"),
    ]:
        assert wrapped_root_default[arg].text == "Cheaper by the Dozen"

    # Subelements in the `isbn` namespace should still be accessible by explicitly
    # giving that prefix or URI
    for arg in [("isbn", "number"), f"{{{isbn_uri}}}number", (isbn_uri, "number")]:
        assert wrapped_root_default[arg].text == "1568491379"

    # Returned ElementWrappers for subelements in a different namespace should be
    # initialized in that other namespace
    wrapped_number: ElementWrapper = wrapped_root_default["isbn", "number"]
    assert wrapped_number._with_prefix == "isbn"

    # Factory for ElementWrappers in the other namespace
    isbn_EWF = ElementWrapperFactory(isbn_prefix)

    # The wrapped element (the root) is in the default namespace, but its methods
    # accessing subelements will use the `isbn` namespace, unless another namespace
    # prefix is explicitly given
    wrapped_root_isbn = isbn_EWF(root)
    assert wrapped_root_isbn._with_prefix == "isbn"

    # Its subelement in the isbn namespace, "isbn:number", can be accessed by
    # giving "number" alone, though supplying the prefix or URI explicitly is still fine
    for arg in ["number", ("isbn", "number"), "isbn:number", f"{{{isbn_uri}}}number"]:
        assert wrapped_root_isbn[arg].text == "1568491379"

    # Subelements in other namespaces can only be accessed by explicitly providing
    # a prefix or URI
    for arg in [("", "title"), f"{{{default_uri}}}title", (default_uri, "title")]:
        assert wrapped_root_isbn[arg].text == "Cheaper by the Dozen"

    # Returned ElementWrappers are initialized with the correct prefixes
    assert wrapped_root_isbn["", "title"]._with_prefix == ""

    # Sanity check that the tree has not been modified by __getitem__
    xml_after = ET.tounicode(tree)
    assert xml_before == xml_after


def test_elementwrapper_getitem_default_namespace_redeclared():
    """Functionality of __getitem__ when a subtree declares a new default namespace."""
    parser = ET.XMLParser(remove_blank_text=True)
    tree = ET.parse(data_dir / "default_namespace.xml", parser=parser)
    root = tree.getroot()
    xml_before = ET.tounicode(tree)

    # Factory for ElementWrappers in the original default namespace
    default_EWF = ElementWrapperFactory("")
    wrapped_notes_default = default_EWF(root)["notes"]

    # The <{urn:loc.gov:books}notes> element has a subtree with a new default namespace:
    #   <p xmlns='http://www.w3.org/1999/xhtml'>Book is <i>funny!</i></p>
    # The paragraph may be accessed by explicitly giving the expanded name in Clark
    # Notation: "{http://www.w3.org/1999/xhtml}p"
    p_default_uri = "http://www.w3.org/1999/xhtml"
    wrapped_p_new_default = wrapped_notes_default[f"{{{p_default_uri}}}p"]
    assert wrapped_p_new_default.text == "Book is "
    assert wrapped_p_new_default._with_prefix == ""

    # Its subelement <i> is in the new default namespace. Since the ElementWrapper
    # for <p> will use <p>'s nsmap, its subelement can be accessed without providing
    # the new default namespace explicitly
    for arg in ["i", ("", "i"), (p_default_uri, "i"), f"{{{p_default_uri}}}i"]:
        wrapped_i_new_default = wrapped_p_new_default["i"]
        assert wrapped_i_new_default.text == "funny!"
        assert wrapped_i_new_default._with_prefix == ""

    # Sanity check that __getitem__ has not modified the tree
    xml_after = ET.tounicode(tree)
    assert xml_before == xml_after


def test_elementwrapper_delitem_default_namespace():
    """Functionality of __deltem__ to delete default-namespaced subelements."""
    tree = ET.parse(data_dir / "default_namespace.xml", parser=parser)
    root = tree.getroot()

    # Default namespace prefix
    default_prefix = ""

    # ElementWrapper for the root, with the default namespace preferred
    default_EWF = ElementWrapperFactory(default_prefix)
    root_default = default_EWF(root)

    # Delete <isbn:number>
    del root_default["isbn", "number"]

    # Delete <p xmlns='http://www.w3.org/1999/xhtml'>Book is <i>funny!</i></p>
    del root_default["notes"]["{http://www.w3.org/1999/xhtml}p"]

    # Delete the <div>
    del root_default["notes"]["p"]["div"]

    # Compare with expected XML file
    expected_tree = ET.parse(
        data_dir / "default_namespace_after_deletions.xml", parser=parser
    )
    assert ET.tounicode(tree) == ET.tounicode(expected_tree)


def test_elementwrapper_setitem_existing_subelements():
    """Basic functionality of __setitem__ to modify text of existing subelements."""
    html = "<html><head/><body><h1>Heading</h1><p>Hello!</p><hr/></body></html>"
    html_root = ET.XML(html)
    html_tree = html_root.getroottree()

    wrapped_root = ElementWrapper(html_root, None)

    wrapped_body = wrapped_root["body"]
    wrapped_p = wrapped_body["p"]
    wrapped_body["p"] = "Goodbye!"
    assert wrapped_body["p"].text == "Goodbye!"
    assert wrapped_p._element.text == "Goodbye!"

    edited_html = ET.tounicode(html_tree)
    assert (
        edited_html
        == "<html><head/><body><h1>Heading</h1><p>Goodbye!</p><hr/></body></html>"
    )


def test_elementwrapper_setitem_missing_subelements():
    """Using __setitem__ to create a subelement with given text."""
    html = "<html><head/><body><h1>Heading</h1><p></p><hr/></body></html>"
    html_root = ET.XML(html)
    html_tree = html_root.getroottree()

    wrapped_root = ElementWrapper(html_root, None)

    # `<p></p>` has no text or subelement at this point
    wrapped_p = wrapped_root["body"]["p"]

    # Inserts a subelement `<b>Bold text!</b>`
    wrapped_p["b"] = "Bold text!"
    assert wrapped_p["b"].text == "Bold text!"
    wrapped_b = wrapped_p["b"]
    assert wrapped_b.text == "Bold text!"
    assert wrapped_b._element.text == "Bold text!"

    # Confirm the tree as a whole has the expected change from <p/> to
    #   <p><b>Bold text!</b></p>
    # and only that change
    edited_html = ET.tounicode(html_tree)
    expected_html = (
        "<html><head/><body><h1>Heading</h1>"
        "<p><b>Bold text!</b></p>"
        "<hr/></body></html>"
    )
    assert edited_html == expected_html


def test_getattr():
    """Use __getattr__ to access an attribute of the underlying Element."""
    html = (
        "<html><head/><body>"
        '<a href="example.org">'
        '<img src="example.gif" alt="Alt text"/>'
        "</a></body></html>"
    )
    html_root = ET.XML(html)
    html_tree = html_root.getroottree()

    wrapped_root = ElementWrapper(html_root, None)
    wrapped_a = wrapped_root["body"]["a"]

    # An ElementWrapper has no method `values()` but its Element does
    a_values = list(wrapped_a.values())
    assert a_values == ["example.org"]
    also_a_values = list(wrapped_a._element.values())
    assert also_a_values == ["example.org"]

    # Similarly, call the `img` Element's method
    #   get(self, key, default=None)
    # for the keys 'src' and 'alt'
    src = wrapped_a["img"].get("src")
    assert src == "example.gif"
    alt_text = wrapped_a["img"].get("alt")
    assert alt_text == "Alt text"

    # Sanity check that this has not modified the tree
    html_out = ET.tounicode(html_tree)
    assert html == html_out



"""Helper class for working with lxml ETrees and Elements.

This is Reinventing The Wheel as a personal exercise -- if this were Real or Serious,
i.e. anyone else was ever going to work with it, I would have just used lxml.Objectify,
which is a robust, time-tested, much-better-designed take on an API for XML trees that
resembles normal Python objects.
(https://lxml.de/objectify.html)

(I thought "why can't I interact with the tree like indexing a dictionary / accessing an
attribute?", and then it turned out that lxml already had that, but I wanted to write my
own implementation anyway. Mostly so I had a chance to write more dunder methods than I
usually have reason to, haha.)
"""
from __future__ import annotations

import functools
import re
from typing import Optional

from lxml import etree as ET

# Underscores present because those are the class names in lxml.etree
Element = ET._Element
ElementTree = ET._ElementTree

# Regex matching names of the form
#   "{namespace}localname"
clark_notation = re.compile(r"\{\S*\}\S*")
# extract "namespace", "localname" as groups 1 and 2
clark_notation_groups = re.compile(r"\{(\S*)\}(\S*)")

# Sentinel indicating "use the ElementWrapper's prefix attribute"
_USE_WRAPPER_NAMESPACE = object()
# Alias for an empty string, which for ElementWrapper is the 'prefix' of a
# default namespace
_USE_DEFAULT_NAMESPACE = ""
# (lxml represents the prefix of a default namespace as None, but for the wrapper's
# purposes, being given a None prefix in a (None, "localname") tuple corresponds to
# No Namespace, while simply not being given a namespace ("localname") indicates
# "use the wrapper's namespace"
_USE_NO_NAMESPACE = None


def try_unpack(func):
    """Replace the original method's first argument with a different representation.

    This decorator is applied to methods whose first (non-self) argument is the name of
    a subelement to be accessed. This subelement name may be namespaced, but may or
    may not have been given with a particular namespace or namespace prefix. The
    original argument is replaced with a corresponding QName if needed.

    All other args and kwargs are passed along as given.

    The subelement name may have been given in the following forms:

    "localname"
        Single string with no namespace specified. The ElementWrapper's preferred
        namespace will be used with the given localname to create a qualified name
        (QName) corresponding to a subelement with that localname in that namespace.

    (None, "localname")
        A prefix of None indicates that no namespace should be used. This may have been
        passed explicitly because the ElementWrapper's prefix attribute specifies a
        namespace.

    ("prefix", "localname")
        Tuple of prefix and localname. These are unpacked and used to create a
        QName replacing the original argument.

    ("uri", "localname")
        Tuple of namespace and localname. Again, these are unpacked and turned into an
        equivalent QName.

    "prefix:localname"
        Single string with a prefix separating the namespace prefix and localname. The
        prefix and localname will are used to create a QName for the subelement.

    "{uri}localname"
        Single string with the namespace URI itself given in braces, i.e.  "Clark
        Notation." This form is unambiguous and may be used as-is.
    """

    @functools.wraps(func)
    def wrapper(self, subelement_name, *args, **kwargs):
        name = subelement_name
        # This would be more Pythonic as a try/except attempting to unpack
        #
        #   prefix, name = subelement_name
        #
        # and catching `ValueError: too many values to unpack (expected 2)` when
        # subelement_name is just a string, and not actually a tuple of (prefix, name).
        #
        # Unfortunately, this will unpack a string of length 2.
        # E.g. for subelement_name = 'ab', this would result in
        #
        #   prefix, name = 'a', 'b'
        #
        # instead of raising and then catching a ValueError like it would for a string
        # of any other length != 2.
        #
        # So, unfortunately, the more Pythonic Way has different behavior which is
        # undesirable here. :(
        if isinstance(subelement_name, tuple):
            # Unpack the name if it is a (prefix, localname), (uri, localname), or
            # (None, localname) pair
            prefix_or_uri, name = subelement_name
            name = self._QName(name, prefix_or_uri)
        elif clark_notation.fullmatch(name):
            # An expanded name given in Clark notation may be used as-is
            pass
        elif ":" in name:
            # If the given name is of the form "prefix:localname", obtain the
            # equivalent QName
            # (Namespace URIs may contain colons, but prefixes cannot, so names
            # of the form prefix:localname will contain only one colon)
            prefix, name = name.split(":", maxsplit=1)
            name = self._QName(name, prefix)
        else:
            # Otherwise, with no prefix specified, the ElementWrapper's own prefix
            # attribute will be used
            name = self._QName(name, _USE_WRAPPER_NAMESPACE)

        return func(self, name, *args, **kwargs)

    return wrapper


class ElementWrapper:
    """TODO."""

    def __init__(self, element: Element, with_prefix):
        """Initialization."""
        self.__dict__["_element"] = element
        if with_prefix is None and None in element.nsmap:
            # If None is in the element's nsmap, the None indicates a default namespace
            self.__dict__["_with_prefix"] = _USE_DEFAULT_NAMESPACE
        elif with_prefix is None and None not in element.nsmap:
            # Otherwise, None indicates using no namespace
            self.__dict__["_with_prefix"] = _USE_NO_NAMESPACE
        else:
            self.__dict__["_with_prefix"] = with_prefix

    @try_unpack
    def __getitem__(self, subelement_name: str | ET.QName) -> "ElementWrapper":
        """Index into a given subelement."""
        found = self._get_or_create_subelement(self._element, subelement_name)
        return ElementWrapper(found, found.prefix)

    @try_unpack
    def __setitem__(self, subelement_name, value):
        """Set the (text) value of a given subelement."""
        self._set_subelement_text(self._element, subelement_name, str(value))
        pass

    @try_unpack
    def __delitem__(self, subelement_name: str | ET.QName) -> None:
        """Remove a given subelement."""
        found = self._maybe_get_subelement(self._element, subelement_name)
        if found is not None:
            self._element.remove(found)
        pass

    @try_unpack
    def iterfind(self, subelement_name: str | ET.QName) -> list["ElementWrapper"]:
        """Find subelements matching a given name."""
        results = self._element.iterfind(subelement_name)
        return [ElementWrapper(found, found.prefix) for found in results]

    @try_unpack
    def create_subelement(self, subelement_name: str | ET.QName) -> "ElementWrapper":
        """Create a new subelement with the given name, even if others already exist."""
        created = ET.SubElement(self._element, subelement_name)
        return ElementWrapper(created, created.prefix)

    @try_unpack
    def __contains__(self, subelement_name) -> bool:
        """Return if there is a subelement with the given name (i.e. `name in self`)."""
        return self._maybe_get_subelement(self._element, subelement_name) is not None

    def __getattr__(self, attr):
        """Retrieve attributes of the underlying Element object.

        If the Element object has no such attribute, try to access the value of
        the XML element for the given attribute name, i.e. `my_elementwrapper.attrib`
        evaluates to "Value" for an element <tagname attrib="Value"/>.
        """
        try:
            return self._element.__getattribute__(attr)
        except AttributeError:
            return self._element.get(key=attr)

    def __setattr__(self, attr, value):
        """Set attribute of the underlying Element object.

        If the Element object has no such attribute, set the value of
        the XML element for the given attribute name, i.e.
        `my_elementwrapper.attrib = value` sets <tagname attrib="value"/>.
        """
        if attr in self._element or attr == "text":
            self._element.__setattr__(attr, value)
        else:
            self._element.set(key=attr, value=str(value))

    def remove_attribute(self, attr):
        """Remove attribute from the element if present."""
        if attr in self._element.attrib:
            self._element.attrib.pop(attr)
        pass

    def _get_subelement(self, element, subelement_name: ET.QName | str) -> Element:
        """Get the wrapped element's subelement.

        Args:
            subelement_name (ET.QName | str):
                Name of the subelement to find.

        Returns:
            Element: the found subelement.

        Raises:
            ValueError: If the subelement is not found.
        """
        found = element.find(str(subelement_name), self._nsmap())
        if found is None:
            raise ValueError(
                f"Could not find required subelement: {str(subelement_name)}."
            )
        return found

    def _set_subelement_text(
        self, element: Element, subelement_name: str | ET.QName, text: str
    ) -> Element:
        """Update the text of a specified entry's subelement.

        If no subelement with the specified name is found, add the subelement before
        setting its text.

        Args:
            entry (Element):
                Parent XML element whose subelement is to be updated.
            subelement_name (str | ET.QName):
                Name of the subelement to update.
            text (str):
                Text to be enclosed by the specified subelement.

        Returns:
            Element: the modified (possibly newly created) subelement.
        """
        subelement = self._get_or_create_subelement(element, subelement_name)
        # if subelement is None:
        #     subelement = ET.SubElement(element, subelement_name)

        subelement.text = text
        return subelement

    def _maybe_get_subelement(
        self, element: Element, subelement_name: ET.QName | str
    ) -> Optional[Element]:
        """Get a specified element's subelement.

        The only reason this one-line method exists is to avoid the visual noise of
        repeatedly providing `self._nsmap` as an argument to Element's `find`.
        Otherwise, `find` would be used directly each time, instead of calling this.

        Args:
            entry (Element):
                Parent XML element whose subelement is to be found.
            subelement_name (ET.QName | str):
                Name of the subelement to find.

        Returns:
            Optional[Element]: the found subelement, or None if not found.
        """
        return element.find(str(subelement_name), self._nsmap())

    def _default_subelement_text(
        self, element: Element, subelement_name: str | ET.QName, text: str
    ) -> Element:
        """Set the text of a specified entry's subelement only if no text is present."""
        subelement = self._get_or_create_subelement(element, subelement_name)
        subelement.text = subelement.text or text
        return subelement

    def _get_or_create_subelement(
        self, element: Element, subelement_name: ET.QName | str
    ) -> Element:
        """Get a specified element's subelement, or add it if not found.

        Args:
            entry (Element):
                Parent XML element whose subelement is to be found.
            subelement_name (ET.QName | str):
                Name of the subelement to find.

        Returns:
           Element: the found or created subelement.
        """
        subelement = self._maybe_get_subelement(element, subelement_name)
        if subelement is not None:
            return subelement
        else:
            # Create missing subelement
            return ET.SubElement(element, subelement_name)

    def _QName(self, name: str, prefix_or_uri=_USE_WRAPPER_NAMESPACE) -> ET.QName | str:
        """Creates qualified name of a given tag in the `reruns` namespace."""
        if prefix_or_uri is _USE_WRAPPER_NAMESPACE:
            # Use the wrapper's namespace setting, possibly No Namespace
            prefix_or_uri = self._with_prefix
        if prefix_or_uri is _USE_NO_NAMESPACE:
            # Return the unqualified name
            return name
        if prefix_or_uri in self._nsmap():
            # Convert the prefix to a URI before using
            return ET.QName(self._nsmap()[prefix_or_uri], name)
        elif prefix_or_uri in self._nsmap().values():
            # Given a namespace URI
            return ET.QName(prefix_or_uri, name)
        else:
            raise ValueError(
                "Argument `prefix_or_uri` could not be determined to be a namespace "
                f"prefix or URI: '{prefix_or_uri}' not found in keys or values of "
                f"namespace map {self._nsmap()}."
            )

    def _nsmap(self) -> dict[str, str]:
        """Get dictionary of XML namespaces to use with `find()`, `findall()`, etc.

        The 'cleaning' is due to the representation of a Default Namespace
        in the `element.nsmap` dictionary -- if there is a default namespace, its
        key is `None` rather than an empty string.

        This makes Mypy find its keys to have type `Optional[str]` instead of `str`,
        which unfortunately makes it believe the dictionary is incompatible with the
        type signature of the Element methods `find()`, `findall()`, etc.

        For the purposes of `ElementWrapper`, a namespace prefix of None indicates
        'use the ElementWrapper's preferred prefix', not 'use the xml element's Default
        Namespace'.
        """
        return self._clean_nsmap(self._element.nsmap)

    def _clean_nsmap(self, nsmap: dict[Optional[str], str]) -> dict[str, str]:
        """Replace a `None` key an with empty string if encountered."""
        return {(k or _USE_DEFAULT_NAMESPACE): v for k, v in nsmap.items()}

    def clear_xml_base(self) -> None:
        """Reset xml:base to an empty string if a base URI is declared or inherited."""
        if self._element.base:
            self._element.base = ""
        pass


class ElementWrapperFactory:
    """Create ElementWrappers with a given preferred prefix."""

    def __init__(self, with_prefix):
        """Initialization."""
        self._with_prefix = with_prefix

    def __call__(self, element: Element | ElementWrapper) -> ElementWrapper:
        """Return an ElementWrapper with the factory's preferred prefix."""
        if isinstance(element, ElementWrapper):
            return ElementWrapper(element._element, self._with_prefix)
        return ElementWrapper(element, self._with_prefix)



"""Rebroadcast old RSS/Atom items to a new feed, in shuffled or chronological order."""
# For the sake of Python versions < 3.10 (for stdlib generics and | for type unions)
from __future__ import annotations

from .feedmodifier import FeedModifier

__all__ = ["FeedModifier"]
